/**
 * class FScore
 *
 * Copyright (C) 2006 Sebastian Pado 
 * Copyright (C) 2012 Eva Mujdricza-Maydt
 *
 * This software is provided under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 */
package de.pado.sigf;

import java.util.*;
import java.io.*;

/**
 * Compute Precision, Recall, or F-Score (beta=1) for observations. 
 * Each observation is a triple of longs (long[3]): 
 * <ul>
 * <li>number of true positives </li>
 * <li>total number of positives according to model </li>
 * <li>total number of positives according to gold standard</li>
 * </ul>
 */
class FScore implements Statistic<Triple<Long>> {

	Long true_pos;      // all true positives
	Long model_pos;     // all positives assigned by the model
	Long gold_pos;      // all positives in the gold standard
	Long no;            // number of observations
	static String measure;		// evaluation measure for the observations: 
								// precision, recall, or f1-measure

	/** Computes the perfomance for the observations. If no evaluation
	 * measure given, f1-measure will be computed as default.
	 * 
	 * @return the requested evaluation score (precision, recall, or f1-measure)
	 */
	@Override
	public double compute() {
		double score;
		switch (measure.charAt(0)) {
			case 'p':
				score = computePrec();
				break;
			case 'r':
				score = computeRec();
				break;
			case 'f':
				score = computeFMeas();
				break;
			default:
				score = computeFMeas();
				break;
		}
		return score;
	}
	
	private double computePrec() {
		if (model_pos == 0) {
			return 0.0;
		}
		return ((double) true_pos / model_pos);
	}
	
	private double computeRec() {
		if (gold_pos == 0) {
			return 0.0;
		} 
		return ((double) true_pos / gold_pos);
	}
	
	private double computeFMeas() {
		double precision = computePrec();
		double recall = computeRec();
		if ((precision == 0) && (recall == 0)) {
			return 0;
		} 
		return (2 * precision * recall) / (precision + recall);
	}

	@Override
	public void add_observation(Triple<Long> observation) {
		true_pos += observation.first();
		model_pos += observation.second();
		gold_pos += observation.third();
		no += 1;
	}

	@Override
	public FScore clone() {
		return new FScore(true_pos, model_pos, gold_pos, no, measure);
	}

	FScore() {
		true_pos = new Long(0);
		model_pos = new Long(0);
		gold_pos = new Long(0);
		no = new Long(0);
		
	}

	FScore(long tp, long mp, long gp, long num_obs, String meas) {
		true_pos = tp;
		model_pos = mp;
		gold_pos = gp;
		no = num_obs;
		measure = new String(meas);
	}

	@Override
	public long num_observations() {
		return no;
	}
	
	public static void setMeasure(String m) {
		measure = m;
	}

}

/**
 * Computation of F-Scores for two datasets and comparison with a randomization
 * test. This class extends the (abstract) ApproximateRandomizationTest class. 
 * It only implements main(), readInput(), and calls the FScore class.
 *
 * @author pado
 *
 */
public class FScoreART extends ApproximateRandomizationTest<Triple<Long>> {

	/**
	 * Read a vector of triples from a file.
	 */
	@Override
	public Vector<Triple<Long>> readInput(String filename) throws IOException, FileNotFoundException {

		Vector<Triple<Long>> v = new Vector<Triple<Long>>();
		BufferedReader fileReader = new BufferedReader(new FileReader(filename));
		String line;
		while ((line = fileReader.readLine()) != null) {
			if (line.length() == 0) {
				continue;
			}
			StringTokenizer linet = new StringTokenizer(line, " ");
			Long true_pos = Long.valueOf(linet.nextToken());
			Long model_pos = Long.valueOf(linet.nextToken());
			Long gold_pos = Long.valueOf(linet.nextToken());

			Triple<Long> triple = new Triple<Long>(true_pos, model_pos, gold_pos);
			v.addElement(triple);
		}
		return v;
	}


	/**
	 * Computation of Precision, Recall, or F-Scores for two datasets 
	 * and comparison with a randomization test.
	 *
	 * @param args {@literal <filename1>, <filename2>, <number of iterations>, 
	 *				<evaluation score>}
	 * @throws ClassNotFoundException
	 * @throws InstantiationException
	 * @throws IllegalAccessException
	 */
	public static void main(String[] args) throws ClassNotFoundException, InstantiationException, IllegalAccessException {
		if (args.length < 3) {
			System.err.println("Usage: FScoreART <file1> <file2> <no. of iterations> <evaluation measure>"
				+ "\nmeasure alternatives: precision, recall, f1-measure");
			System.exit(1);
		}
		ApproximateRandomizationTest<Triple<Long>> art = new FScoreART();

		String measure = null; 
		if (args.length > 3) {
			char m = args[3].toLowerCase().charAt(0);
			switch (m) {
				case 'p':
					measure = "precision";
					break;
				case 'r':
					measure = "recall";
					break;
				case 'f':
					measure = "f1-measure";
					break;
				default:
					measure = "f1-measure"; //default scoring is f1-measure
					break;
			}
		}
		FScore.setMeasure(measure);
		art.run(args[0], args[1], FScore.class, Long.valueOf(args[2]).longValue(), measure);
	}
}
