/**
 * CRFalign: A CRF-based sentence aligner for parallel corpora.
 * 
 * Copyright (C) 2012  Eva Mujdricza-Maydt
 * Copyright (C) 2012  Huiqin Körkel-Qu
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 * 
 * Developers:
 * - Eva Mujdricza-Maydt (mujdricz@cl.uni-heidelberg.de)
 * - Huiqin Körkel-Qu (huiqin@cl.uni-heidelberg.de)
 */

package annotators.champollion;

import com.cybozu.labs.langdetect.LangDetectException;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileWriter;
import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;

/**Preprocessing entails writing a bash script which starts
 * the Champollion aligner for each file-pair.
 * <p/>
 * NOTE that in the lib folder we provide a set of files for the German-English 
 * language pair. For details read readme in the champollion_GE subfolder.
 *
 * @author Eva Mujdricza-Maydt
 * @lastChange 20120626
 */
class Champollion_Preprocessing {
	
	public static void main(String[] args)
		throws FileNotFoundException, IOException, LangDetectException {
		try {
			writeBashScript_forChampollion(args[0], args[1], args[2], args[3], args[4], args[5]);
		} catch (ArrayIndexOutOfBoundsException e) {
			System.out.println("Usage: The preprocessing expects the following command line arguments:\n"
				+ "- <path to Champollion>\n- <name of the current corpus>\n"
				+ "<path to the current data set>\n- <extension of the >\n"
				+ "<output script file name with path>");
		}
	}
	
	/** Writes a bash script for starting Champollion for all file-pairs.
	 * <p/>
	 * NOTE that the german text should be lemmatized already, the english text
	 * should be only tokenized.
	 * 
	 * @param champollionPath path to Champollion Tool Kit
	 * @param corpusPath path to the current corpus
	 * @param extension_src main.Constants_Main.EXT_TEXT_LEMMATA
	 * @param extension_trg main.Constants_Main.EXT_TEXT_TOKENS
	 * @param outputScriptName name of the bash script with path
	 * @throws FileNotFoundException
	 * @throws IOException
	 * @throws LangDetectException 
	 */
	private static void writeBashScript_forChampollion(
		String corpusPath, String extension_src, String extension_trg,
		String champollionPath, String champollionModel,
		String outputScriptName)
			throws FileNotFoundException, IOException {
		
		System.out.println("corpusPath: " + corpusPath);
		System.out.println("champollionPath: " + champollionPath);
		System.out.println("champollionModel: " + champollionModel);
		System.out.println("output script: " + outputScriptName);

		ArrayList<String> fileStems_src = utils.Helper.extractFileNamesByEnding_woEnding(
			corpusPath + main.Constants_Main.FOLDER_SRC, extension_src);
		ArrayList<String> fileStems_trg = utils.Helper.extractFileNamesByEnding_woEnding(
			corpusPath + main.Constants_Main.FOLDER_SRC, extension_src);
		//check
		if (fileStems_src.size() != fileStems_trg.size()) {
			String msg = "\nSource and target side do not contain the equal number of input files.\n";
			throw new ArrayIndexOutOfBoundsException(msg);
		}
		for (String fs : fileStems_src) {
			if (!fileStems_trg.contains(fs)) {
				String msg = "\nSource file name \"" + fs + "\" cannot be found in the target file name list.\n";
				throw new FileNotFoundException(msg);
			}
		}
		
		System.out.println("\nNumber of input file pairs: " + fileStems_src.size() + " source documents");

		HashMap<String, String> srcToTrgInputMap = new HashMap<String, String>();
		HashMap<String, String> srcInputToChampollionOutputMap = new HashMap<String, String>();
		for (String fileStem : utils.Helper.sortAL_str(fileStems_src)) {
			
			String srcInput = main.Constants_Main.FOLDER_SRC + fileStem + extension_src;
			System.out.println("srcInput: " + srcInput);

			String trgInput = main.Constants_Main.FOLDER_TRG + fileStem + extension_trg;
			System.out.println("trgInput: " + trgInput);
			
			String champollionOutput = main.Constants_Main.FOLDER_ALIGNMENTS + fileStem + main.Constants_Main.EXT_CHAMPOLLIONALIGNMENT;
			System.out.println("champollionOutput: " + champollionOutput);
			
			srcToTrgInputMap.put(srcInput, trgInput);
			srcInputToChampollionOutputMap.put(srcInput, champollionOutput);
		}

		writeChampollionBashScript(srcToTrgInputMap, srcInputToChampollionOutputMap, 
			champollionPath, champollionModel, corpusPath, outputScriptName);
		System.out.println("-> written bash script: " + outputScriptName);

	}
	
	/** Writes a bash script for starting Champollion for one file-pair.
	 * <p/>
	 * NOTE that the german text should be lemmatized already, the english text
	 * should be only tokenized.
	 * 
	 * @param srcToTrgInputFiles map of the file pair names
	 * @param srcInputToChampollionOutputMap map of the input file names to the 
	 *			corresponding output champollion-alignment file names
	 * @param champollionPath path to Champollion Tool Kit, e.g.
	 *			/pathTo/champollion/champollion-1.1
	 * @param champollionModel the name of the model to use (we provide a model
	 *			for German-English language pair in the lib folder)
	 * @param currDataPath path to the current data set
	 * @param bashFileName the name of the output file with path
	 * @throws IOException 
	 */
	private static void writeChampollionBashScript(
		HashMap<String, String> srcToTrgInputFiles, //german: tokenized+lemmatized text, english: tokenized text
		HashMap<String, String> srcInputToChampollionOutputMap,
		String champollionPath,
		String champollionModel,
		String currDataPath,
		String bashFileName) 
		throws IOException {
		
		//this replacing is only needed in case of a mixture of windows-style
		//path and unix-style path
		bashFileName = bashFileName.replace("\\", main.Constants_Main.SLASH);
		
		currDataPath = currDataPath.replace("\\", main.Constants_Main.SLASH);
		if (currDataPath.endsWith(main.Constants_Main.SLASH)) {
			currDataPath = currDataPath.substring(0, currDataPath.length()-1);
		}
		
		champollionPath = champollionPath.replace("\\", main.Constants_Main.SLASH);
		if (champollionPath.endsWith(main.Constants_Main.SLASH)) {
			champollionPath = champollionPath.substring(0, champollionPath.length()-1);
		}

		FileWriter fw = new FileWriter(new File(bashFileName));
		
		fw.write("#!/bin/bash\n");
		fw.write("CTK=" + champollionPath + "\n");
		fw.write("export CTK\n");
		fw.write("PATH_CURRDATA=" + currDataPath + "\n");
		fw.write("PATH_CHAMPOLLION=" + champollionPath + "\n");
		fw.write("MODEL_CHAMPOLLION=" + champollionModel + "\n");
		fw.write("#\n#\n");
		int i = 1;
		int all = srcToTrgInputFiles.size();
		for (String in : utils.Helper.sortKeys_str(srcToTrgInputFiles.keySet())) {
			String out = srcToTrgInputFiles.get(in);
			String champOut = srcInputToChampollionOutputMap.get(in);

			//normalize fileNames
			in = in.replace("\\", main.Constants_Main.SLASH);
			out = out.replace("\\", main.Constants_Main.SLASH);

			fw.write("echo \"" + (i++) + "/" + all + ": " + champOut + "\"\n");
			
			//one line in the script e.g. 
			// /path/to/champollion/champollion-1.1$ ./bin/champollion.GE /path/to/4135819_1of1-4133743_1of1.text_lemmata /path/to/4135819_1of1-4133743_1of1.text_tokens /path/to/4135819_1of1-4133743_1of1.champollionAlign
			fw.write("$PATH_CHAMPOLLION" + main.Constants_Main.SLASH + "bin" + main.Constants_Main.SLASH
				+ "$MODEL_CHAMPOLLION"
				+ " " + currDataPath + main.Constants_Main.SLASH + in
				+ " " + currDataPath + main.Constants_Main.SLASH + out
				+ " " + currDataPath + main.Constants_Main.SLASH + champOut
				+ "; echo \"\"\n");
		}
		fw.write("echo \"\"\n");
		fw.write("echo \"DONE\"\n");
		fw.close();
	}
	
}
