/**
 * CRFalign: A CRF-based sentence aligner for parallel corpora.
 * 
 * Copyright (C) 2012  Eva Mujdricza-Maydt
 * Copyright (C) 2012  Huiqin Körkel-Qu
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 * 
 * Developers:
 * - Eva Mujdricza-Maydt (mujdricz@cl.uni-heidelberg.de)
 * - Huiqin Körkel-Qu (huiqin@cl.uni-heidelberg.de)
 */

package annotators.languageDetector;

import com.cybozu.labs.langdetect.Detector;
import com.cybozu.labs.langdetect.DetectorFactory;
import com.cybozu.labs.langdetect.LangDetectException;
import java.io.File;
import java.io.FileReader;
import java.io.IOException;


/** The OpenSubtitles corpus sometimes annotates texts with incorrect language;
 * to filter out texts with incorrect language annotation, we use 
 * the LanguageDetector by Nakatani Shuyo. 
 * <p>
 * http://code.google.com/p/language-detection/wiki/Tutorial
 * <br/>
 * Copyright (c) 2010-2011 Cybozu Labs, Inc. All rights reserved.
 * <br/>
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with the License. You may obtain a copy of the License at 
 * http://www.apache.org/licenses/LICENSE-2.0
 * <br/>
 * Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the specific language governing permissions and limitations under the License. 
 * <p/>
 * 
 * @author Eva Mujdricza-Maydt
 * @lastChange 20120803
 */
public class DetectLanguage {
	
	/**
	 * Below, you can read some parts of the
	 * tutorial to the program:
	 *  * <br/>
	* http://code.google.com/p/language-detection/wiki/Tutorial
	* <br/>
	* Copyright (c) 2010-2011 Cybozu Labs, Inc. All rights reserved.
	* <br/>
	* Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with the License. You may obtain a copy of the License at 
	* http://www.apache.org/licenses/LICENSE-2.0
	* <br/>
	* Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the specific language governing permissions and limitations under the License. 
	* <p/>
Initialize library<br/>
------------------<br/>
Before using this library, call DetectorFactory#loadProfile() once to initialize.<br/>

DetectorFactory.loadProfile(profileDirectory);<br/>

The parameter of this method is a directory which has files of language profiles. The language profiles are bundled with this library, so specify "trunk/profile" in repository as the parameter of loadProfile().
Set target text Detector class is an interface of this language detection library. This class can be only instanced via DetectorFactory class.<br/>
<p/>
Due to set target text to detect language, use Detector#append() method.<br/>
------------------<br/>
{@code Detector detector = DetectorFactory.create();}<br/>
{@code detector.append(text);}

The Detector#append() method has 2 versions of parameter. One is to set whole text with java.lang.String parameter or another is to set text stream with java.io.Reader parameter. Use whichever suit to your application.<br/>

Create a new Detector instance if you need to detect again.
<p/>
Retrieve detection result<br/>
------------------<br/>
There are 2 ways of retriving result after setting target text.<br/>
Detector#detect<br/>

The Detector#detect method returns a single language name which has the most probability. Language names are represented by language codes. See also LanguageList .<br/>

{@code Detector detector = DetectorFactory.create();}<br/>
{@code detector.append(text);}<br/>
{@code String lang = detector.detect();}<br/>

Detector#getProbabilities<br/>

The Detector#getProbabilities method returns a languages list with their probabilities. The list is in order of probability.<br/>

{@code Detector detector = DetectorFactory.create();}<br/>
{@code detector.append(text);}<br/>
{@code ArrayList<Language> langlist = detector.getProbabilities();}<br/>
	 */

	/** detects the language of the text.
	 * 
	 * @param documentName name of the file with plain text
	 * @param profileDirectory path to the profile files of the detector
	 * @return the language name abbreviation string detected for the file
	 * @throws LangDetectException
	 * @throws IOException 
	 */
	public static String detectLanguageOfDocument(
		String documentName, String profileDirectory) 
		throws LangDetectException, IOException {
		
		try {
			DetectorFactory.loadProfile(profileDirectory);
		} catch (LangDetectException lde) {
			//nothing to do, because the profile is already loaded
		}
		
		Detector detector = DetectorFactory.create();
		detector.append(new FileReader(new File(documentName)));
		
		return detector.detect();
		
		//other possible way:
//		System.out.println("Language: " + lang);
//		ArrayList<Language> langlist = detector.getProbabilities();
//		for (Language l : langlist) {
//			System.out.println("\t- " + l.toString());
//		}	
//		return langlist.get(0).lang;
	}
	
}
