/**
 * CRFalign: A CRF-based sentence aligner for parallel corpora.
 * 
 * Copyright (C) 2012  Eva Mujdricza-Maydt
 * Copyright (C) 2012  Huiqin Körkel-Qu
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 * 
 * Developers:
 * - Eva Mujdricza-Maydt (mujdricz@cl.uni-heidelberg.de)
 * - Huiqin Körkel-Qu (huiqin@cl.uni-heidelberg.de)
 */

package annotators.treetagger;

import com.cybozu.labs.langdetect.LangDetectException;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.FileWriter;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.util.ArrayList;
import java.util.HashMap;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.parsers.SAXParser;
import javax.xml.parsers.SAXParserFactory;
import javax.xml.stream.XMLOutputFactory;
import javax.xml.stream.XMLStreamException;
import javax.xml.stream.XMLStreamWriter;
import objects.Sentence;
import objects.Word;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;
import org.xml.sax.helpers.DefaultHandler;

/** We use TreeTagger to lemmatize the german and english texts.
 *
 * @author Eva Mujdricza-Maydt
 * @lastChange 20120626
 */
public class ExtractCorpus_forTreeTagger {
	
	/** list for the sentences in the current text file */
	static ArrayList<Sentence> sentencesInDoc;
	

	/** Input for TreeTagger in xml-format.
	 * 
	 * @param pathToDataSet path to the current data set (training or testing data)
	 * @param fileStems list of file names without their file extension (We 
	 *			assume, that the source and target files are named equal.)
	 * @param encoding encoding for the xml-file. NOTE that for TreeTagger, 
	 *			the encoding should be ISO-8859-1.
	 * @throws FileNotFoundException
	 * @throws IOException
	 * @throws LangDetectException
	 * @throws XMLStreamException
	 * @throws ParserConfigurationException
	 * @throws SAXException 
	 */
	public static void writeTreeTaggerInput_xml(
			String pathToDataSet, ArrayList<String> fileStems, 
			String encoding)
			throws FileNotFoundException, IOException, LangDetectException, XMLStreamException, ParserConfigurationException, SAXException {

		for (String fileStem : utils.Helper.sortAL_str(fileStems)) {
			
			String inputFileName = fileStem + main.Constants_Main.EXT_TEXT_TOKENS;
			ArrayList<Sentence> sentences = corpora.ExtractCorpus.readInPlainText(
				pathToDataSet + main.Constants_Main.FOLDER_SRC, inputFileName, true);

			String outputFileName = pathToDataSet + main.Constants_Main.FOLDER_SRC + fileStem + Constants_TreeTagger.EXT_TREETAGGER_INPUT_XML;
			writeTreeTaggerInput_xml(sentences, outputFileName, encoding);
			System.out.print(".");
		}
		
		for (String fileStem : utils.Helper.sortAL_str(fileStems)) {
			String inputFileName = fileStem + main.Constants_Main.EXT_TEXT_TOKENS;
			ArrayList<Sentence> sentences = corpora.ExtractCorpus.readInPlainText(
				pathToDataSet + main.Constants_Main.FOLDER_TRG, inputFileName, true);

			String outputFileName = pathToDataSet + main.Constants_Main.FOLDER_TRG + fileStem + Constants_TreeTagger.EXT_TREETAGGER_INPUT_XML;
			writeTreeTaggerInput_xml(sentences, outputFileName, encoding);
			System.out.print(".");
		}
		System.out.println();
	}
	
	
	/** Writes the xml-format input file for TreeTagger.
	 * NOTE that xml-format is needed, if the texts could entail sentences
	 * with no words (empty sentence). In this case, the empty sentence in 
	 * the plain-format input would be not considered by TreeTagger.
	 * 
	 * @param sentences list of {@code Sentence} objects for the given file
	 * @param outputFileName name of the output xml-file with path
	 * @param encoding the encoding for the output xml-file. 
	 *			NOTE that for TreeTagger, the encoding should be ISO-8859-1.
	 * @throws IOException
	 * @throws FileNotFoundException
	 * @throws XMLStreamException 
	 */
	private static void writeTreeTaggerInput_xml(
		ArrayList<Sentence> sentences,
		String outputFileName, String encoding)
		throws IOException, FileNotFoundException, XMLStreamException {

		//check
		utils.Helper.checkFileForWriting(outputFileName);

		XMLOutputFactory factory = XMLOutputFactory.newInstance();
		FileOutputStream output = new FileOutputStream(outputFileName);
		XMLStreamWriter writer = factory.createXMLStreamWriter(
			output, encoding);
		
		//<?xml version="1.0" encoding="ISO-8859-1"?>
		writer.writeStartDocument(encoding, "1.0");
		writer.writeCharacters("\n");

		writer.writeStartElement(Constants_TreeTagger.ELEMENT_sentences);
		writer.writeCharacters("\n");

		for (Sentence sent : sentences) {
			writer.writeStartElement(Constants_TreeTagger.ELEMENT_s);
			writer.writeAttribute(Constants_TreeTagger.ATTRIBUTE_id, String.valueOf(sent.getSentenceNr()));
			writer.writeCharacters("\n");
			try {
				ArrayList<Word> words = sent.getWordObjList();
				for (Word w : words) {
					writer.writeCharacters(w.getWord() +  "\n");
				}
			} catch (NullPointerException npe) {
				System.out.println("\n! " + outputFileName + ": " + npe.getMessage());
			}
			writer.writeEndElement();
			writer.writeCharacters("\n");
		}

		writer.writeEndElement();
		writer.writeCharacters("\n");
		writer.writeEndDocument();
		writer.writeCharacters("\n");

		writer.flush();
		writer.close();

		output.close();
	}

	
	/** We starts TreeTagger with two bash scripts (one for the source files,
	 * one for the target files, respectively) which entail
	 * a separate call for all files.
	 * 
	 * @param pathToDataSet path to the current data set (training or testing data)
	 * @param fileStems list of file names without their file extension (We 
	 *			assume, that the source and target files are named equal.)
	 * @param pathToTreeTagger path to the TreeTagger
	 * @param pathToScripts path to the scripts folder which serves as output
	 *			folder for the bash scripts generated with this method
	 * @return the written bash script file names with path (in order that they
	 *			can be executed later)
	 * @throws FileNotFoundException
	 * @throws IOException
	 * @throws LangDetectException 
	 */
	public static ArrayList<String> writeBashScriptsForTreeTagger_xml(
		String pathToDataSet, ArrayList<String> fileStems, 
		String pathToTreeTagger,
		String pathToScripts) 
			throws FileNotFoundException, IOException, LangDetectException {
		
		ArrayList<String> scriptNames = new ArrayList<String>(2);
		
		String treetaggerPath = pathToTreeTagger;
		String options = Constants_TreeTagger.OPTIONS_TREETAGGER;

	
		HashMap<String, String> ttInOut_src = new HashMap<String, String>();
		for (String fileStem:  utils.Helper.sortAL_str(fileStems)) {
			String inputFileName = fileStem + Constants_TreeTagger.EXT_TREETAGGER_INPUT_XML;
			String outputFileName = fileStem + Constants_TreeTagger.EXT_TREETAGGER_OUTPUT_XML;
			ttInOut_src.put(inputFileName, outputFileName);
		}

		String language_src = getSourceLanguageForTreeTagger();
		String bashFileName_src = Constants_TreeTagger.EXTPART_TREETAGGER_BASHSCRIPT + language_src + main.Constants_Main.EXT_SH;
		String bashFileNameWP_src = pathToScripts + bashFileName_src;
		writeTreeTaggerBashScript(
				ttInOut_src, treetaggerPath, pathToDataSet + main.Constants_Main.FOLDER_SRC, language_src, options, bashFileNameWP_src);
		System.out.println("-> written bash script: " + bashFileNameWP_src);
		scriptNames.add(bashFileName_src);

		HashMap<String, String> ttInOut_trg = new HashMap<String, String>();
		for (String fileStem : utils.Helper.sortAL_str(fileStems)) {
			String inputFileName = fileStem + Constants_TreeTagger.EXT_TREETAGGER_INPUT_XML;
			String outputFileName = fileStem + Constants_TreeTagger.EXT_TREETAGGER_OUTPUT_XML;
			ttInOut_trg.put(inputFileName, outputFileName);
		}
		String language_trg = getTargetLanguageForTreeTagger();
		String bashFileName_trg = Constants_TreeTagger.EXTPART_TREETAGGER_BASHSCRIPT + language_trg + main.Constants_Main.EXT_SH;
		String bashFileNameWP_trg = pathToScripts + bashFileName_trg;
		writeTreeTaggerBashScript(
				ttInOut_trg, treetaggerPath, pathToDataSet + main.Constants_Main.FOLDER_TRG, language_trg, options, bashFileNameWP_trg);
		System.out.println("-> written bash script: " + bashFileNameWP_trg);
		scriptNames.add(bashFileName_trg);
		
		return scriptNames;
	}

	/** Writes a bash script which can call the TreeTagger for a list of files.
	 * 
	 * @param ttInOut map with input and output file names
	 * @param treetaggerPath path to the TreeTagger
	 * @param currDataPath path to the current data set (training/test data)
	 * @param language language of the texts. For german, use "german", for 
	 *			english, use "english"
	 * @param options the command line arguments to use for TreeTagger
	 * @param bashFileName output bash script file name with path
	 * @throws IOException 
	 */
	private static void writeTreeTaggerBashScript(
		HashMap<String, String> ttInOut,
		String treetaggerPath,
		String currDataPath,
		String language,
		String options,
		String bashFileName) throws IOException  {

		//normalize paths
		treetaggerPath = treetaggerPath.replace("\\", main.Constants_Main.SLASH);
		currDataPath = currDataPath.replace("\\", main.Constants_Main.SLASH);
		bashFileName = bashFileName.replace("\\", main.Constants_Main.SLASH);

		if (treetaggerPath.endsWith(main.Constants_Main.SLASH)) {
			treetaggerPath = treetaggerPath.substring(0, treetaggerPath.length()-1);
		}
		
		if (currDataPath.endsWith(main.Constants_Main.SLASH)) {
			currDataPath = currDataPath.substring(0, currDataPath.length()-1);
		}

		FileWriter fw = new FileWriter(new File(bashFileName));

		fw.write("#!/bin/bash\n");
		fw.write("PATH_TREETAGGER=" + treetaggerPath + "\n");
		fw.write("PATH_CURRDATA=" + currDataPath + "\n");
		fw.write("PATH_LIB=$PATH_TREETAGGER" + main.Constants_Main.SLASH + "lib\n");
		fw.write("LNG=" + language + "\n");
		fw.write("PARFN=$PATH_LIB" + main.Constants_Main.SLASH + "$LNG.par\n");
		fw.write("#\n#\n");

		int all = ttInOut.size();
		int i = 1;
		for (String in : utils.Helper.sortKeys_str(ttInOut.keySet())) {
			String out = ttInOut.get(in);

			//normalize fileNames
			in = in.replace("\\", main.Constants_Main.SLASH);
			out = out.replace("\\", main.Constants_Main.SLASH);

			fw.write("echo \"" + (i++) + "/" + all + ": " + in + "\"\n");
			fw.write("$PATH_TREETAGGER" + main.Constants_Main.SLASH + "bin" + main.Constants_Main.SLASH + "tree-tagger " 
				+ options
				+ " $PARFN $PATH_CURRDATA"
				+ main.Constants_Main.SLASH + in
				+ " $PATH_CURRDATA" + main.Constants_Main.SLASH + out
				+ "; echo \"\"\n");
		}
		fw.write("echo \"\"");
		fw.close();
	}
	
	/** Reads the xml-format output of the TreeTagger in a list of 
	 * {@code Sentence} objects.
	 * 
	 * @param inputFileName output of the TreeTagger
	 * @return the list of the sentences in the output, with part-of-speech and
	 *			lemma information provided by TreeTagger
	 * @throws IOException
	 * @throws ParserConfigurationException
	 * @throws SAXException 
	 */
	public static ArrayList<Sentence> readInTreeTaggerOutput_xml(
			String inputFileName)
			throws IOException,
			ParserConfigurationException, SAXException {

		try {
			SAXParserFactory factory = SAXParserFactory.newInstance();
			SAXParser saxParser = factory.newSAXParser();
			DefaultHandler handler = new XMLHandler_treeTaggerOutput();

			try {
				File f = new File(inputFileName);

				InputStream inputStream = new FileInputStream(f);
				String encoding = "ISO-8859-1"; //NOTE that TreeTagger should have ISO-8859-1 encoding
				InputSource is = new InputSource(new InputStreamReader(inputStream, encoding));
				is.setEncoding(encoding);

				saxParser.parse(is, handler);

			} 
			catch (IOException ioex) {
				System.err.println("File " + inputFileName + " not found.");
				ioex.printStackTrace();
				System.exit(2);
			}

		} catch (ParserConfigurationException pcex) {
			pcex.printStackTrace();
			System.exit(2);
		} catch (SAXException saxex) {
			saxex.printStackTrace();
			System.exit(2);
		} 
		return sentencesInDoc;
	}
	

	/** For now, german and english are available as languages.
	 * 
	 * @return the name of the language of the source files (with the original
	 *			setting, this will be "german")
	 */
	private static String getSourceLanguageForTreeTagger() {
		String lang = null;
		if (main.Constants_Main.LANG_SRC.equals(main.Constants_Main.LANG_DE)) {
			lang = "german";
		} else if (main.Constants_Main.LANG_SRC.equals(main.Constants_Main.LANG_EN)) {
			lang = "english";
		} else {
			String msg = "\nNo source language set for tree-tagger (" + main.Constants_Main.LANG_SRC + ")\n";
			throw new NullPointerException(msg);
		}
		return lang;
	}

	/** For now, german and english are available as languages.
	 * 
	 * @return the name of the language of the target files (with the original
	 *			setting, this will be "english")
	 */
	private static String getTargetLanguageForTreeTagger() {
		String lang = null;
		if (main.Constants_Main.LANG_TRG.equals(main.Constants_Main.LANG_DE)) {
			lang = "german";
		} else if (main.Constants_Main.LANG_TRG.equals(main.Constants_Main.LANG_EN)) {
			lang = "english";
		} else {
			String msg = "\nNo target language set for tree-tagger (" + main.Constants_Main.LANG_TRG + ")\n";
			throw new NullPointerException(msg);
		}
		return lang;
	}
	
}
