/**
 * CRFalign: A CRF-based sentence aligner for parallel corpora.
 * 
 * Copyright (C) 2012  Eva Mujdricza-Maydt
 * Copyright (C) 2012  Huiqin Körkel-Qu
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 * 
 * Developers:
 * - Eva Mujdricza-Maydt (mujdricz@cl.uni-heidelberg.de)
 * - Huiqin Körkel-Qu (huiqin@cl.uni-heidelberg.de)
 */
 
package corpora;

import com.cybozu.labs.langdetect.LangDetectException;
import java.io.BufferedReader;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.util.ArrayList;
import javax.xml.parsers.ParserConfigurationException;
import objects.Alignment_DocLevel;
import objects.Alignment_SentLevel;
import objects.Sentence;
import objects.Word;
import org.xml.sax.SAXException;

/** Corpus extraction.
 *
 * @author Eva Mujdricza-Maydt
 * @lastChange 20120626
 */
public class ExtractCorpus {

	/** alignments for one file pair */
	public static Alignment_DocLevel corpusAlignment_oneDoc; //from the alignment-file (e.g. de-en.xml.gz)
	/** name of the current text file */
	public static String srcFileNameForAligning; //from the alignment-file (e.g. de-en.xml.gz)

	/**
	 * Reads plain text (with tokens or lemmata) one sentence per line.
	 *
	 * @param pathToFile	path to the file name we want to read (NOTE that 
	 *			this path is not equals to the path for a data set, but it is 
	 *			a subpath of them.)
	 * @param fileName	name of the file (pathToFile+fileName should be a valid
	 * path to this file)
	 * @param whetherCountEmtpyLinesAsSentence	if {@code true}, then also empty lines
	 * will be treaten as sentences. otherwise, empty lines will be just ignored
	 * @return	list of sentences in the file
	 * @throws FileNotFoundException
	 * @throws IOException
	 */
	public static ArrayList<Sentence> readInPlainText(
		String pathToFile, String fileName, boolean whetherCountEmtpyLinesAsSentence)
		throws FileNotFoundException, IOException {

		ArrayList<Sentence> sentences = new ArrayList<Sentence>();
		BufferedReader br = new BufferedReader(new FileReader(new File(pathToFile + fileName)));
		StringBuilder sb = new StringBuilder();
		int c = 0;
		while ((c = br.read() ) != -1) {
			sb.append((char)c);
		}
		sb.append(" "); //!!! This is needed to get a possible last empty line in the file.
		String[] lines = sb.toString().split("\n");

		int count = 1;
		for (String line : lines) {
			if ((!whetherCountEmtpyLinesAsSentence)  && (line.trim().isEmpty())) {
				continue;
			}
		
			String[] lineArr = line.trim().split(" ");
			ArrayList<Word> wordList = new ArrayList<Word>();
			for (int i = 0; i < lineArr.length; i++) {
				Word w = new Word();
				w.setSentenceNr(count);
				w.setWord(lineArr[i]);
				w.setWordInSentNr(i + 1);
				wordList.add(w);
			}

			Sentence sent = new Sentence();
			sent.setSentenceNr(count++);
			sent.setWordList(wordList);
			sentences.add(sent);
		}
		br.close();
		return sentences;
	}
	

	/** Writes one sentence per line with lemmata. The lemma information will 
	 * be read from the TreeTagger output for the files.
	 * 
	 * @param pathToDataSet path to the current data set
	 * @param fileStems list of the file names without their extension
	 * @throws FileNotFoundException
	 * @throws IOException
	 * @throws ParserConfigurationException
	 * @throws SAXException
	 * @throws LangDetectException 
	 */
	public static void writePlainTexts_lemmata_fromTreeTaggerOutput(
		String pathToDataSet,
		ArrayList<String> fileStems)
		throws FileNotFoundException, IOException,
		ParserConfigurationException, SAXException, LangDetectException {
		
		for (String fileStem : utils.Helper.sortAL_str(fileStems)) {

			String treeTaggerOutputFile_src = pathToDataSet + main.Constants_Main.FOLDER_SRC + fileStem + annotators.treetagger.Constants_TreeTagger.EXT_TREETAGGER_OUTPUT_XML;
			ArrayList<Sentence> textInSentences_src = annotators.treetagger.ExtractCorpus_forTreeTagger.readInTreeTaggerOutput_xml(treeTaggerOutputFile_src);
			String plainTextFileName_src_lemma = pathToDataSet + main.Constants_Main.FOLDER_SRC + fileStem + main.Constants_Main.EXT_TEXT_LEMMATA;
			writePlainText_lemmata(textInSentences_src, plainTextFileName_src_lemma);
			System.out.print(".");
		
			String treeTaggerOutputFile_trg = pathToDataSet + main.Constants_Main.FOLDER_TRG + fileStem + annotators.treetagger.Constants_TreeTagger.EXT_TREETAGGER_OUTPUT_XML;
			ArrayList<Sentence> textInSentences_trg = annotators.treetagger.ExtractCorpus_forTreeTagger.readInTreeTaggerOutput_xml(treeTaggerOutputFile_trg);
			String plainTextFileName_trg_lemma = pathToDataSet + main.Constants_Main.FOLDER_TRG + fileStem + main.Constants_Main.EXT_TEXT_LEMMATA;
			writePlainText_lemmata(textInSentences_trg, plainTextFileName_trg_lemma);
			System.out.print(".");
		}
		System.out.println();
	}

	
	/** Writes one sentence per line with the tokens of the sentences.
	 * 
	 * @param textInSentences list of sentences in the file
	 * @param plainTextFileName output file name with path
	 * @throws FileNotFoundException
	 * @throws IOException 
	 */
	public static void writePlainText_tokens(
		ArrayList<Sentence> textInSentences,
		String plainTextFileName) throws FileNotFoundException, IOException {

		//check
		utils.Helper.checkFileForWriting(plainTextFileName);

		FileWriter fw = new FileWriter(new File(plainTextFileName));

		for (int i = 0; i < textInSentences.size(); i++) {
			if (i > 0) {
				fw.write("\n");
			}
			Sentence s = textInSentences.get(i);
			ArrayList<Word> words = s.getWordObjList_suppressException();

			StringBuilder sb = new StringBuilder();
			for (Word w : words) {
				sb.append(w.getWord()).append(" ");
			}
			fw.write(sb.toString().trim());
		}

		fw.close();
	}

	/**
	 * Writes one sentence per line with lemmata of the tokens in the sentences.
	 *
	 * @param textInSentences list of sentences in the file
	 * @param plainTextFileName output file name with path
	 * @throws FileNotFoundException
	 * @throws IOException
	 */
	private static void writePlainText_lemmata(
		ArrayList<Sentence> textInSentences,
		String plainTextFileName) throws FileNotFoundException, IOException {

		//check
		utils.Helper.checkFileForWriting(plainTextFileName);

		FileWriter fw = new FileWriter(new File(plainTextFileName));

		for (int i = 0; i < textInSentences.size(); i++) {
			if (i > 0) {
				fw.write("\n");
			}
			Sentence s = textInSentences.get(i);
			ArrayList<Word> words = s.getWordObjList_suppressException();

			StringBuilder sb = new StringBuilder();
			for (Word w : words) {
				sb.append(w.getLemma()).append(" ");
			}
			fw.write(sb.toString().trim());
		}

		fw.close();
	}

	/** Returns the sentence alignment if it is in the given list of sentence
	 * alignments, otherwise null.
	 * 
	 * @param sentenceAlignments list of given sentence alignments
	 * @param candidate sentence number searched for in the alignments
	 * @param srcOrTrgSide there are two allowed values: "src", if the sentence 
	 *			number will be searched on the source side; 
	 *			"trg" if the sentence number will be 
	 *			searched on the target side of the alignment
	 * @return the found sentence-level alignment with the candidate sentence 
	 *			number, or {@code null} if no appropriate alignment found in the
	 *			given alignment list
	 */
	public static Alignment_SentLevel containsSentAlign(
		ArrayList<Alignment_SentLevel> sentenceAlignments,
		int candidate, String srcOrTrgSide) {

		Alignment_SentLevel found = null;

		if (srcOrTrgSide.equals("src")) {
			for (Alignment_SentLevel al : sentenceAlignments) {
				ArrayList<Integer> nrs = al.getSourceSentAlign();
				if (nrs.contains(candidate)) {
					found = al;
					break;
				}
			}
		} else if (srcOrTrgSide.equals("trg")) {
			for (Alignment_SentLevel al : sentenceAlignments) {
				ArrayList<Integer> nrs = al.getTargetSentAlign();
				if (nrs.contains(candidate)) {
					found = al;
					break;
				}
			}
		} else {
			String msg = "\nNeeded third argument: \"src\" or \"trg\"\n";
			throw new IllegalArgumentException(msg);
		}

		return found;
	}
}
