/**
 * CRFalign: A CRF-based sentence aligner for parallel corpora.
 * 
 * Copyright (C) 2012  Eva Mujdricza-Maydt
 * Copyright (C) 2012  Huiqin Körkel-Qu
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 * 
 * Developers:
 * - Eva Mujdricza-Maydt (mujdricz@cl.uni-heidelberg.de)
 * - Huiqin Körkel-Qu (huiqin@cl.uni-heidelberg.de)
 */
package main;

import com.cybozu.labs.langdetect.LangDetectException;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.util.*;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.stream.XMLStreamException;
import org.xml.sax.SAXException;

/** Starts the aligner.
 *
 * @author Eva Mujdricza-Maydt
 */
public class Main {
	
	/** Starts the predefined processes for aligning parallel files on the 
	 * sentence-level. 
	 * 
	 * @param args the command line arguments. Each process has a defined set of 
	 *			required command line argument list.
	 */
	public static void main(String[] args) 
		throws IOException, FileNotFoundException, LangDetectException, 
		ParserConfigurationException, SAXException, InterruptedException, XMLStreamException {

		if (args.length == 0) {
			printProcessList();
			System.exit(2);
		}
		if (args.length == 1 && 
			(args[0].toLowerCase().equals("h") || args[0].toLowerCase().equals("help")
			|| args[0].toLowerCase().equals("-h") || args[0].toLowerCase().equals("-help")
			|| args[0].toLowerCase().equals("--h") || args[0].toLowerCase().equals("--help"))
			) {
			printProcessList();
			System.exit(2);
		}
	
		String process = args[0];
		System.out.println("PROCESS: " + process);
		if (!getParameter().containsKey(process.toLowerCase())) {
			printProcessesWithParameterList(process);
		}
		
		if (process.toLowerCase().equals(main.Constants_Main.PROCESS.preproc.name().toLowerCase())) {
		
			if (args.length < (getParameter().get(main.Constants_Main.PROCESS.preproc.name()).size())) {
				printProcessesWithParameterList(main.Constants_Main.PROCESS.preproc.name());
				System.exit(2);
			}
				
			String pathToDataSet = utils.Helper.normalizePath(args[1]); //e.g. /path/to/gutenberg/train/
			String originalFileExtension = args[2]; //e.g. .txt
			String pathToScripts = utils.Helper.normalizePath(args[3]); //e.g. /path/to/scripts/
			String pathToTreeTagger = utils.Helper.normalizePath(args[4]); //e.g. /resources/processors/tagger/tree-tagger/
			
			System.out.println("path to the current data set: " + pathToDataSet);
			System.out.println("original file extension:      " + originalFileExtension);
			System.out.println("path to scripts:              " + pathToScripts);
			System.out.println("path to the TreeTagger:       " + pathToTreeTagger);
			System.out.println();
			
			ArrayList<String> fileStems = utils.Helper.extractFileNamesByEnding_woEnding(
				pathToDataSet + main.Constants_Main.FOLDER_SRC, originalFileExtension);
			
			System.out.println("There are " + fileStems.size() + " file pairs in the corpus path\n"
				+ pathToDataSet + "\n");
			
			System.out.println("\nWRITING TOKENS TEXT (" + main.Constants_Main.EXT_TEXT_TOKENS + "):");
			corpora.gutenberg.ExtractGutenberg.writeGutenbergPlainTexts_tokens(
				pathToDataSet, fileStems, originalFileExtension);

			System.out.println("\nWRITING TREETAGGER INPUT (" + annotators.treetagger.Constants_TreeTagger.EXT_TREETAGGER_INPUT_XML + "):");
			annotators.treetagger.ExtractCorpus_forTreeTagger.writeTreeTaggerInput_xml(
				pathToDataSet, fileStems, annotators.treetagger.Constants_TreeTagger.encoding);

			System.out.println("\nWRITING TREETAGGER BASH-SCIRPT (" + annotators.treetagger.Constants_TreeTagger.EXTPART_TREETAGGER_BASHSCRIPT + "<LANGUAGE>" + main.Constants_Main.EXT_SH + "):");
			ArrayList<String> bashscriptFileNames = 
				annotators.treetagger.ExtractCorpus_forTreeTagger.writeBashScriptsForTreeTagger_xml(
				pathToDataSet, fileStems, pathToTreeTagger, pathToScripts);

			System.out.println("\nSTARTING TREETAGGER BASH-SCIRPT (" + annotators.treetagger.Constants_TreeTagger.EXTPART_TREETAGGER_BASHSCRIPT + "<LANGUAGE>" + main.Constants_Main.EXT_SH + "):");
			for (String bash : bashscriptFileNames) {
				utils.Helper.startBashScript(
					pathToTreeTagger,
					pathToScripts, 
					bash, null, false, true, null);
			}

			System.out.println("\nWRITING LEMMATA TEXT (" + main.Constants_Main.EXT_TEXT_LEMMATA + "):");
			corpora.ExtractCorpus.writePlainTexts_lemmata_fromTreeTaggerOutput(
				pathToDataSet, fileStems);
				
		}

		else if (process.toLowerCase().equals(main.Constants_Main.PROCESS.html.name().toLowerCase().toLowerCase())) {
			
			if (args.length < (getParameter().get(main.Constants_Main.PROCESS.html.name()).size())) {
				printProcessesWithParameterList(main.Constants_Main.PROCESS.html.name());
				System.exit(2);
			}

			String pathToDataSet = utils.Helper.normalizePath(args[1]);
			String alignmentExtension = args[2];
			String outputPath = utils.Helper.normalizePath(args[3]);
			main.Constants_Main.EXT_OTHEROPUSFORMATALIGNMENT = alignmentExtension;
			
			System.out.println("path to the current data set: " + pathToDataSet);
			System.out.println("alignment file extension:     " + alignmentExtension);
			System.out.println("output path:                  " + outputPath);
			System.out.println();
			
			corpora.ExtractCorpus_forEvaluation.writeParallelHtmlFiles(
				pathToDataSet, alignmentExtension, 
				annotators.treetagger.Constants_TreeTagger.EXT_TREETAGGER_OUTPUT_XML, outputPath);

		}

		else if (process.toLowerCase().equals(main.Constants_Main.PROCESS.eval.name().toLowerCase())) {
			
			if (args.length < (getParameter().get(main.Constants_Main.PROCESS.eval.name()).size())) {
				printProcessesWithParameterList(main.Constants_Main.PROCESS.eval.name());
				System.exit(2);
			}
			
			String pathDataSet = utils.Helper.normalizePath(args[1]);
			String extensionGoldenAlignment = args[2];
			String extensionSystemAlignment = args[3];
			String outputFileNameWithPath = args[4];
		
			System.out.println("path to the current data set:          " + pathDataSet);
			System.out.println("golden alignment file extension:       " + extensionGoldenAlignment);
			System.out.println("alignment file extension to evaluate : " + extensionSystemAlignment);
			System.out.println("output file name:                      " + outputFileNameWithPath);
			System.out.println();

			//default
			boolean takeOneToOne = true;
			boolean takeManyToMany = true;
			boolean takeZero = true;

			eval.EvaluateCorpus.evaluateAlignments_forDocs(
				pathDataSet,
				extensionGoldenAlignment, //against this annotation
				extensionSystemAlignment, //annotation to evaluate
				takeOneToOne, takeManyToMany, takeZero,
				outputFileNameWithPath);
		}
		
		else if (process.toLowerCase().equals(main.Constants_Main.PROCESS.crftrain.name().toLowerCase())) {
			
			if (args.length < (getParameter().get(main.Constants_Main.PROCESS.crftrain.name()).size())) {
				printProcessesWithParameterList(main.Constants_Main.PROCESS.crftrain.name());
				System.exit(2);
			}
			
			String pathDataSet = utils.Helper.normalizePath(args[1]);
			String alignmentExtension = args[2];
			String patternFileName = args[3];
			String modelFileName = args[4];
		
			System.out.println("path to the current data set: " + pathDataSet);
			System.out.println("alignment file extension:     " + alignmentExtension);
			System.out.println("name of the pattern file:     " + patternFileName);
			System.out.println("name of the model file:       " + modelFileName);
			System.out.println();

			String[] trainArgs = new String[args.length-1];
			for (int i = 1; i < args.length; i++) {
				trainArgs[i-1] = args[i];
			}
			
			generalTrain.crfTrain.main(trainArgs);

		}
		
		else if (process.toLowerCase().equals(main.Constants_Main.PROCESS.crftest.name().toLowerCase())) {
			
			if (args.length < (getParameter().get(main.Constants_Main.PROCESS.crftest.name()).size())) {
				printProcessesWithParameterList(main.Constants_Main.PROCESS.crftest.name());
				System.exit(2);
			}
		
			String pathDataSet = utils.Helper.normalizePath(args[1]);
			String pathTrainingProc = utils.Helper.normalizePath(args[2]);
			String extension = args[3];
			String modelFileName = args[4];
		
			System.out.println("path to the current data set:        " + pathDataSet);
			System.out.println("path to the training data set:       " + pathTrainingProc);
			System.out.println("extension for the output alignments: " + extension);
			System.out.println("name of the output model:            " + modelFileName);
			System.out.println();

			String[] testArgs = new String[args.length-1];
			for (int i = 1; i < args.length; i++) {
				testArgs[i-1] = args[i];
			}
			generalTest.crfTest.main(testArgs);

		}

		else if (process.toLowerCase().equals(main.Constants_Main.PROCESS.signiftest.name().toLowerCase())) {
			
			if (args.length < (getParameter().get(main.Constants_Main.PROCESS.signiftest.name()).size())) {
				printProcessesWithParameterList(main.Constants_Main.PROCESS.signiftest.name());
				System.exit(2);
			}
			
			String pathToDataSet = utils.Helper.normalizePath(args[1]);
			String libPath = utils.Helper.normalizePath(args[2]);

			String pathToScripts = utils.Helper.normalizePath(args[3]);
			int rounds = Integer.parseInt(args[4]); //10000 is a good default value
			String measure = args[5];
			String outputFileName = utils.Helper.normalizePath(args[6]);
			
			System.out.println("path to the current data set:  " + pathToDataSet);
			System.out.println("number of iterations:          " + rounds);
			System.out.println("measure (p/r/f):               " + measure);
			System.out.println("significance output file name: " + outputFileName);
			
			String goldExtension = args[7];
			System.out.println("extension of gold alignment files: " + goldExtension);
			System.out.println("extension of other alignment files:");
			int firstIndex = 8;
			String[] extensions = new String[args.length-firstIndex];
			for (int i = firstIndex; i < args.length; i++) {
				extensions[i-firstIndex] = args[i];
				System.out.println("- " + args[i]);
			}
			System.out.println();
			
			eval.Significance.testSignificance(pathToDataSet, libPath, pathToScripts,
				rounds, measure, goldExtension, extensions, outputFileName);
		}
		
		
		else if (process.toLowerCase().equals(main.Constants_Main.PROCESS.convert.name().toLowerCase())) {
			
			if (args.length < (getParameter().get(main.Constants_Main.PROCESS.convert.name()).size())) {
				printProcessesWithParameterList(main.Constants_Main.PROCESS.convert.name());
				System.exit(2);
			}
			
			String pathToDataSet = utils.Helper.normalizePath(args[2]);
			String inputExtension = args[3];
			String outputExtension = args[4];
			System.out.println("path to the current data set:          " + pathToDataSet);
			System.out.println("extension of the input alignments:     " + inputExtension);
			System.out.println("extension of the converted alignments: " + outputExtension);
			System.out.println();
			
			System.out.println("CONVERTING " + args[1] + " alignments into opus-format alignments.");
			if (args[1].toLowerCase().equals("champollion")) {
				annotators.champollion.ExtractChampollionOutput.writeChampollionAlignmentsInOpusFormat(
					pathToDataSet + main.Constants_Main.FOLDER_ALIGNMENTS, inputExtension, outputExtension);
			} else if (args[1].toLowerCase().equals("gargantua")) {
				corpora.gutenberg.ExtractGutenberg.writeGargantuaAlignmentsInOpusFormat(
					pathToDataSet + main.Constants_Main.FOLDER_ALIGNMENTS, inputExtension, outputExtension);
			} else if (args[1].toLowerCase().equals("hunalign")) {
				annotators.hunalign.ExtractHunalignOutput.writeHunalignAlignmentsInOpusFormat(
					pathToDataSet + main.Constants_Main.FOLDER_ALIGNMENTS, inputExtension, outputExtension);
			} else {
				System.err.println("Not provided alignment output convertion with \"" + args[1] + "\"");
				System.exit(1);
			}
				
		}
		
		
		else if (process.toLowerCase().equals(main.Constants_Main.PROCESS.normalizeopus.name().toLowerCase())) {
			
			if (args.length < (getParameter().get(main.Constants_Main.PROCESS.normalizeopus.name()).size())) {
				printProcessesWithParameterList(main.Constants_Main.PROCESS.normalizeopus.name());
				System.exit(2);
			}

			String pathToOpus = utils.Helper.normalizePath(args[1]);
			String yearFrom = args[2];
			String yearTo = args[3];
			String pathToDataSet = utils.Helper.normalizePath(args[4]);
			String libFolderPath = utils.Helper.normalizePath(args[5]);
			
			System.out.println("path to the input OPUS data: " + pathToOpus);
			System.out.println("- first considered year:     " + yearFrom);
			System.out.println("- last considered year:      " + yearTo);
			System.out.println("path to the output data set: " + pathToDataSet);
			System.out.println("path to the \"lib\" folder:  " + libFolderPath);
			System.out.println();
			
			corpora.opensubtitles.ExtractOpensubltitles.normalizeOpusCorpus(
				pathToOpus, yearFrom, yearTo, pathToDataSet, libFolderPath);
		}
		
	}

	
	/** Returns a map of possible command line parameters for the 
	 * possible processes you can start for the sentence aligner.
	 * The first parameter in the command line shows the process you can
	 * start. This parameter is stored as key in the map.
	 * The further parameters are listed in the value of the map.
	 * 
	 * @return Returns a map of possible command line parameters for the 
	 *			possible processes you can start for the sentence aligner.
	 *	
	 */
	private static Map<String, ArrayList<String>> getParameter() {
		
		Map<String, ArrayList<String>> parameters = 
			new HashMap<String, ArrayList<String>>();
		
		ArrayList<String> prep = new ArrayList<String>(5);
		prep.add("<path to the current data set>"); //e.g. /path/to/thiscorpus/train/
		prep.add("<extension of the text files>"); //e.g. .txt
		prep.add("<path for scripts>");//e.g. /path/to/scripts/
		prep.add("<path to the TreeTagger>"); //e.g. /resources/processors/tagger/tree-tagger/
		prep.add("\n\nBrief description: The preprocessing starts the TreeTagger and extracts from the output of the TreeTagger the texts consisting of the lemmata instead of the tokens.");
		parameters.put(main.Constants_Main.PROCESS.preproc.name(),
			prep);
		
		ArrayList<String> html = new ArrayList<String>(4);
		html.add("<path to the current data set>");
		html.add("<extension of the given alignments>"); //the alignments should be in OPUS-style format
		html.add("<output path>"); 
		html.add("\n\nBrief description: This process writes one html-file per parallel file which entails the source-target alignments according to the given alignments. These html-files can make several things easier:\n"
			+ "- You can analyse the errors of the system alignments in a demonstrative way.\n"
			+ "- You can build a golden alignment file based on a baseline alignment file easily, just reading the html file and correct the alignments if the sentence-pairs are not translations from each other.\n"
			+ "NOTE that the given alignment files have to be in the opus-style format (see README).\n"
			+ "NOTE that for missing sentence alignments, zero-alignments will be generated automatically, so that no sentence will be missing in the html-text.");
		parameters.put(main.Constants_Main.PROCESS.html.name(),
			html);
		

		ArrayList<String> crftrain = new ArrayList<String>(4);
		crftrain.add("<path to the training data set>");
		crftrain.add("<extension of the alignment files>");
		crftrain.add("<pattern file name with path>");
		crftrain.add("<output model name with path>");
		crftrain.add("\n\nBrief description: Trains the conditional random fields model with Wapiti according to the training data set.");
		parameters.put(main.Constants_Main.PROCESS.crftrain.name(),
			crftrain);
		
		ArrayList<String> crftest = new ArrayList<String>(4);
		crftest.add("<path to the test data set>");
		crftest.add("<path to the training data set with training procedure outputs>");
		crftest.add("<alignment extension for crf-alignments>");
		crftest.add("<input model name with path>");
		crftest.add("\n\nBrief description: Aligns the test data set with the crf model.");
		parameters.put(main.Constants_Main.PROCESS.crftest.name(),
			crftest);

		
		ArrayList<String> eval = new ArrayList<String>(5);
		eval.add("<path to the current data set>");
		eval.add("<extension of the gold alignment files>");
		eval.add("<extension of the system alignment files>");
		eval.add("<evaluation output file with path>");
		eval.add("\n\nBrief description: This procedure evaluates system alignments against given golden alignments. "
			+ "\nNOTE that both alignments should be placed in the align/ subfolder of the given data set, which is maintained, if the alignments are generated by CRFalign, "
			+ "but you are also allowed to copy (opus-format) alignment files into the align/ subfolder to evaluate them against another alignments. "
			+ "The procedure outputs an evaluation summary on the console, and writes an evaluation file with the evaluation scores precision, "
			+ "recall and f1-measure for each file, and last, for all files together.");
		parameters.put(main.Constants_Main.PROCESS.eval.name(),
			eval);
		
		
		ArrayList<String> signifTest = new ArrayList<String>(6);
		signifTest.add("<path to the test data set>");
		signifTest.add("<path of the folder lib>");
		signifTest.add("<path of the folder srcipts>"); //in- and output path of the temporary files
		signifTest.add("<rounds>");
		signifTest.add("(p|r|f)"); //measure
		signifTest.add("<output file name with path>");
		signifTest.add("<extension of the gold alignment files>");
		signifTest.add("<space-separated list of extensions of alignment files to test>");
		signifTest.add("\n\nBrief description: Computes the pair-wise significance among the given list of alignments with the significance tester by Sebastian Padó (with pace parameter set to 10).  The results will be printed to the console, and written in the given output file.\n"
			+ "NOTE that all of the alignments should be in OPUS-style format (see README.txt).");
		parameters.put(main.Constants_Main.PROCESS.signiftest.name(),
			signifTest);

		ArrayList<String> convert = new ArrayList<String>(5);
		convert.add("(champollion|gargantua|hunalign)");
		convert.add("<path to the current data set>");
		convert.add("<extension of the alignment files>");
		convert.add("<extension of the converted alignment files>");
		convert.add("\n\nBrief description: This process converts the alignment outputs of Gargantua, Champollion, or hunalign into an OPUS-style format (see in README).\n" 
			+ "NOTE that if there is no zero alignments in the original output (such in case of Gargantua), then in the converted output will be no zero alignments either.\n"
			+ "NOTE that the converted version of a hunalign output looks different from the original one, because hunalign has a special alignment notation.");
		parameters.put(main.Constants_Main.PROCESS.convert.name(),
			convert);
		
		
		ArrayList<String> normOpus = new ArrayList<String>(5);
		normOpus.add("<path to the OpenSubtitles2011-corpus data>");
		normOpus.add("<year from>");
		normOpus.add("<year to>");
		normOpus.add("<path to the current data set you need>");
		normOpus.add("<path of the lib folder>");
		normOpus.add("\n\nBrief description: Copies the file pairs for years \"from\" to \"to\" (both inclusive) "
			+ "into the required folder structure for CRFalign (see below). The original file names in the OpenSubtitles2011 (opus) corpus "
			+ "are not the same on the source and target side, so the copied file names will be composed of the source and target file names like this: source.xml, target.xml -> source-target.txt.\n"
			+ "The original xml-files will be extracted to plain text files.\n"
			+ "NOTE that we omit files which do not have a file pair, or which entail html-annotation.\n"
			+ "NOTE that the input downloaded OpenSubtitles2011 data should have the following folder structure:\n"
			+ "/path/to/opensubtitles/\n"
			+ "+--de/ (the de/ subfolder of the decompressed file de.tar.gz)\n"
			+ "+--en/ (like above, the en/ subfolder of the decompressed file en.tar.gz)\n"
			+ "+--de-en.xml.gz");
		parameters.put(main.Constants_Main.PROCESS.normalizeopus.name(),
			normOpus);
		
		return parameters;
	}
	
	/**
	 * Prints all parameter lists for all provided processes.
	 */
	private static void printProcessList() {
		Map<String, ArrayList<String>> parameter = getParameter();
		System.out.println("Call the program with one of the process names as parameter to get more information.");
		for (String it : utils.Helper.sortKeys_str(parameter.keySet())) {
			System.out.println("\t" + it);
		}
		System.out.println();
	}
	
	/** Prints the parameter list for the given process.
	 * 
	 * @param par name of a provided process
	 */
	private static void printProcessesWithParameterList(String par) {
		Map<String, ArrayList<String>> parameters = getParameter();
		if (!parameters.containsKey(par)) {
			System.err.println("NO PROCESS NAME \"" + par + "\" PROVIDED.");
			printProcessList();
			System.exit(2);
		}
		
		StringBuilder sb = new StringBuilder(par);
		for (String p : parameters.get(par)) {
			sb.append(" ").append(p);
		}
		
		System.out.println("\nPARAMETER LIST:\n" + sb.toString());
	}

}
